import importlib
import traceback
import typing
import io
import sys
import enum
import json
import inspect
from IPython.display import display, DisplayHandle

class Biyam:
    @staticmethod
    def _native_to_biyam_type(tp: typing.Type) -> typing.Any:
        if tp is bool:
            return 'Check'
        elif tp in (int, float):
            return 'Number'
        elif tp is str:
            return 'String'
        return None

    @staticmethod
    def procedure(ret_type: typing.Optional[typing.Type]=None,
                  args: typing.Dict[str, typing.Any]=(),
                  inline: typing.Optional[bool]=None,
                  no_func: bool=False,
                  no_proc: bool=False)\
            -> typing.Callable[[typing.Callable], typing.Callable]:
        def decorator(fn: typing.Callable) -> typing.Callable:
            meta: typing.Dict[str, typing.Any] = {}
            no_func_ = no_func
            ret_type_ = ret_type

            if ret_type is None:
                if 'return' in fn.__annotations__:
                    if fn.__annotations__['return'] is None:
                        no_func_ = True
                    ret_type_ = fn.__annotations__['return']

            if ret_type_ is not None:
                rtp = Biyam._native_to_biyam_type(ret_type_)
                if rtp is not None:
                    meta['ret_type'] = rtp

            meta['args'] = []
            for param in inspect.signature(fn).parameters:
                spec = {'name': param}
                tp_ = fn.__annotations__.get(param, None)
                if param in fn.__annotations__ and tp_:
                    spec['type'] = Biyam._native_to_biyam_type(tp_)
                meta['args'].append(spec)

            for name, tp in args.items():
                for spec in meta['args']:
                    if spec['name'] == name:
                        if isinstance(tp, type):
                            spec['type'] = Biyam._native_to_biyam_type(tp)
                        elif isinstance(tp, typing.Iterable) and not isinstance(tp, str):
                            spec['type'] = {i: i for i in tp}
                        else:
                            spec['type'] = tp
                        break

            if inline is not None:
                meta['inline'] = inline
            meta['no_func'] = no_func_
            meta['no_proc'] = no_proc

            setattr(fn, '_biyam_meta_', meta)
            return fn

        return decorator

    @staticmethod
    def _gather_procedures(mod_name: str) -> typing.Dict[str, typing.Any]:
        mod = importlib.import_module(mod_name)

        return {
                k: getattr(v, '_biyam_meta_')
                for k in dir(mod) if not k.startswith('_')
                for v in (getattr(mod, k, None),)
                if hasattr(v, '_biyam_meta_') and hasattr(v, '__call__')
        }

    def __init__(self, *mods: str) -> None:
        self._info = {
            'procedures': {
                mod: Biyam._gather_procedures(mod)
                for mod in mods
            }
        }

    def _ipython_display_(self, **kwargs: typing.Any)\
            -> typing.Optional[DisplayHandle]:
        return display({
            'text/plain': "Biyam Workspace",
            'application/x.biyam.workspace': '',
        }, metadata=self._info, raw=True, **kwargs)



import io
import typing
import importlib
import sys
import traceback
from zmq.eventloop.zmqstream import ZMQStream
from IPython.core.interactiveshell import InteractiveShell
from IPython.core.displaypub import CapturingDisplayPublisher
from IPython.core.displayhook import CapturingDisplayHook
from ipykernel.ipkernel import IPythonKernel


class _CommWriter:
    def __init__(self, kernel: IPythonKernel, caller: str) -> None:
        self._kernel = kernel
        self._caller = caller

    def _send(self, **msg: typing.Any) -> None:
        self._kernel.session.send(self._kernel.iopub_socket, 'biyam', msg)

    def output(self, name: str, text: str) -> None:
        self._send(caller=self._caller,
                   type='output',
                   output={
                       'output_type': 'stream',
                       'name': name,
                       'text': text
                   })

    def resolve(self, ret: typing.Any) -> None:
        self._send(caller=self._caller,
                   type='result',
                   ok=True,
                   ret=ret)

    def reject(self) -> None:
        self._send(caller=self._caller,
                   type='result',
                   ok=False)


class _PseudoWriter(io.TextIOBase):
    def __init__(self, comm_writer: _CommWriter, name: str) -> None:
        self._comm_writer = comm_writer
        self._name = name

    def writable(self) -> bool:
        return True

    def write(self, message: str) -> int:
        self._comm_writer.output(self._name, message)
        return len(message)


class _DummyList(list):
    def __init__(self, comm_writer: _CommWriter) -> None:
        super().__init__()
        self._comm_writer = comm_writer

    def append(self, item: typing.Any) -> None:
        if isinstance(item, typing.Sequence):
            msg = {
                'output_type': 'display_data',
                'data': item[0],
                'metadata': item[1]
            }
        else:
            msg = item
        self._comm_writer.output(msg)


class BiyamRpcManager:
    def __init__(self, msg_type: str, ipy: InteractiveShell) -> None:
        ipy.kernel.shell_handlers[msg_type] = self._handle_biyam
        self._kernel = ipy.kernel
        self._shell = ipy

    def _handle_biyam(self, stream: ZMQStream, ident: typing.List[bytes], msg: typing.Dict[str, typing.Any]) -> None:
        content = msg.get('content')
        caller = content.get('caller', None)
        if caller is None:
            return

        comm = _CommWriter(self._kernel, caller)

        stdout_ = sys.stdout
        stderr_ = sys.stderr
        displayhook_ = sys.displayhook
        shell = self._shell
        display_pub_ = shell.display_pub if shell else None

        try:
            sys.stdout = typing.cast(typing.Any,
                                     _PseudoWriter(comm, 'stdout'))
            sys.stderr = typing.cast(typing.Any,
                                     _PseudoWriter(comm, 'stderr'))
            if shell:
                outputs = _DummyList(comm)
                shell.display_pub = CapturingDisplayPublisher()
                shell.display_pub.outputs = outputs
                sys.displayhook = CapturingDisplayHook(shell, outputs)

            mod_name = content.get('mod', '__main__')

            mod = importlib.import_module(mod_name)
            fn_ = getattr(mod, content.get('fn', 'main'))
            if not hasattr(fn_, '_biyam_meta_'):
                raise PermissionError(fn_.__name__ + ' is not exposed')
            ret = fn_(**content.get('args', {}))

            comm.resolve(ret)
        except Exception:
            traceback.print_exc()
            comm.reject()
        finally:
            sys.stdout = stdout_
            sys.stderr = stderr_
            sys.displayhook = displayhook_
            if shell:
                shell.display_pub = display_pub_



import os
from setuptools import setup
from setuptools.command.build_py import build_py


class JsBuildCommand(build_py):
    def run(self):
        import subprocess
        os.chdir('js')
        try:
            subprocess.check_call(['yarn', 'install'])
            subprocess.check_call(['yarn', 'gulp'])
        finally:
            os.chdir('..')
        super().run()


setup(
        name="biyam",
        version="2017.12.22a0.dev1",
        packages=["biyam"],
        package_data={'biyam': ['*.xml', 'static/*']},
        url='https://github.com/gwangyi/biyam',
        license='MIT',
        author='Sungkwang Lee',
        author_email='gwangyi.kr@gmail.com',
        description='Blockly w/ Python on Jupyter Notebook',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
        ],
        install_requires=['jupyter'],
        cmdclass={'build_py': JsBuildCommand},
        zip_safe=False
)
